Require Import List.
Require Import Models.
Require Import TypeJudge.
Require GenModel.

(* Finally introducing the syntax... *)

Module MakeModel (M : CC_Model).

Include GenModel.MakeModel M.
Import T.
Import Term.

Fixpoint int_trm t :=
  match t with
  | Srt prop => T.prop
  | Srt kind => T.kind
  | Ref n => T.Ref n
  | App u v => T.App (int_trm u) (int_trm v)
  | Abs T M => T.Abs (int_trm T) (int_trm M)
  | Prod T U => T.Prod (int_trm T) (int_trm U)
  end.
Definition int_env := List.map int_trm.

Lemma int_lift : forall n t,
  eq_term (int_trm (lift n t)) (T.lift n (int_trm t)).
Admitted.

Lemma int_subst : forall u t,
  eq_term (int_trm (subst u t)) (T.subst (int_trm u) (int_trm t)).
Admitted.

Lemma int_not_kind : forall T, T <> Srt kind -> int_trm (unmark_app T) <> T.kind.
red; intros.
apply H.
destruct T; try discriminate.
destruct s; trivial; discriminate.
destruct T1; discriminate.
Qed.

Hint Resolve int_not_kind eq_typ_not_kind.


Lemma int_sound : forall e M M' T,
  eq_typ e M M' T ->
  J.typ (int_env (unmark_env e)) (int_trm (unmark_app M)) (int_trm (unmark_app T)) /\
  J.eq_typ (int_env (unmark_env e)) (int_trm (unmark_app M)) (int_trm (unmark_app M')).
induction 1; simpl; intros.
 (* Srt *)
 split.
  apply R.typ_prop.
  apply R.refl.
 (* Ref *)
 split.
  destruct H0.
  subst t.
  unfold lift; rewrite unmark_lift.
  fold (lift (S v) (unmark_app x)); rewrite int_lift.
  apply R.typ_var.
  elim H1; simpl; auto.

  apply R.refl.
 (* Abs *)
 destruct IHeq_typ1.
 clear IHeq_typ2.
 destruct IHeq_typ3.
 split.
  apply R.typ_abs; eauto.
  apply R.eq_typ_abs; trivial.
 (* App *)
 destruct IHeq_typ1.
 destruct IHeq_typ3.
 split.
  rewrite unmark_subst0 with (1:=H2).
  rewrite int_subst.
  apply R.typ_app with (int_trm (unmark_app V)); eauto.
  apply R.eq_typ_app; trivial.
 (* Prod *)
 destruct IHeq_typ1.
 destruct IHeq_typ2.
 split.
  apply R.typ_prod; trivial.
  destruct s2; auto.

  apply R.eq_typ_prod; trivial.
 (* Beta *)
 destruct IHeq_typ1.
 destruct IHeq_typ3.
 split.
  rewrite unmark_subst0 with (1:=H2).
  rewrite int_subst.
  apply R.typ_beta; eauto.

  rewrite unmark_subst0 with (1:=typ_refl2 _ _ _ _ H1).
  rewrite int_subst.
  apply R.eq_typ_beta; eauto.
 (* Red *)
 destruct IHeq_typ1.
 destruct IHeq_typ2.
 split; trivial.
 apply R.typ_conv with (int_trm (unmark_app T)); eauto.
 (* Exp *)
 destruct IHeq_typ1.
 destruct IHeq_typ2.
 split; trivial.
 apply typ_refl2 in H0.
 apply R.typ_conv with (int_trm (unmark_app T')); eauto.
 apply R.sym; trivial.
Qed.

(***********)

Load "template/Library.v".

(*
  Lemma valid_context_ok : forall e,
    valid_context e = true -> forall M M', ~ eq_typ e M M' FALSE.
*)

  Lemma non_provability : forall T,
    (forall x, ~ el (int_trm (unmark_app T)) vnil x) ->
    forall M M', ~ eq_typ nil M M' T.
red in |- *; intros.
apply int_sound in H0.
destruct H0 as (H0,_).
red in H0.
apply H with (int (int_trm (unmark_app M)) vnil).
apply H0.
red; intros.
destruct n; discriminate.
Qed.

End MakeModel.