
Require Import List Models.
Require ZFecc ZFindtypes.
Require ModelZF.
Import ZFgrothendieck.
Import ZF ZFsum ZFnats ZFord ZFfix ZFecc ZFindtypes IZF.

Import ModelZF.CCM.
Import ModelZF.BuildModel.
Import T R J.

Lemma simpl_int_lift : forall i n x T,
  int (lift (S n) T) (V.cons x i) == int (lift n T) i.
intros.
destruct T as [(T,Tm)|]; simpl; auto with *.
apply Tm.
red; red; intros; reflexivity.
Qed.

Lemma lift0_term : forall T, eq_term (lift 0 T) T.
destruct T as [(T,Tm)|]; simpl; trivial.
do 2 red; intros.
apply Tm.
rewrite V.lams0.
red; red; intros; apply H.
Qed.

Lemma simpl_int_lift1 : forall i x T,
  int (lift 1 T) (V.cons x i) == int T i.
intros.
rewrite simpl_int_lift; rewrite lift0_term; reflexivity.
Qed.

Lemma simpl_lift1 : forall i n x y T,
  int (lift1 (S n) T) (V.cons x (V.cons y i)) == int (lift1 n T) (V.cons x i).
intros.
unfold lift1.
do 2 rewrite int_lift_rec_eq.
apply int_morph; auto with *.
rewrite <- V.cons_lams.
 rewrite <- V.cons_lams.
  reflexivity.

  do 2 red; intros.
  rewrite H; reflexivity.

 do 2 red; intros.
 rewrite H; reflexivity.
Qed.

Lemma lift10: forall T, eq_term (lift1 0 T) T.
intros.
unfold lift1.
destruct T as [(T,Tm)|]; simpl; trivial.
do 2 red; intros.
apply Tm.
do 2 red; intros.
destruct a; simpl.
 unfold V.lams; simpl.
 apply H.

 unfold V.lams; simpl.
 unfold V.shift; simpl.
 replace (a-0) with a; auto with arith.
Qed.

(* Projection to ordinal *)

Parameter toOrd : set -> set.
Parameter toOrd_morph : morph1 toOrd.
Parameter toOrd_ord : forall o, isOrd o -> toOrd o == o.
Parameter toOrd_isOrd : forall x, isOrd (toOrd x).
Existing Instance toOrd_morph.

(* Subtyping *)

Definition sub_typ_covariant : forall e U V1 V2,
  sub_typ (U::e) V1 V2 ->
  sub_typ e (Prod U V1) (Prod U V2).
unfold sub_typ; simpl; intros.
apply ZFcoc.cc_prod_covariant with (3:=H1).
 red; red; intros.
 rewrite H3; auto with *.

 red; intros.
 apply H; trivial.
 apply vcons_add_var; trivial.
Qed.

(* Universes *)

Definition type (n:nat) := cst (ecc n).

Lemma typ_Prop : forall e, typ e prop (type 0).
red; intros; simpl.
apply (grot_succ_in gr).
Qed.

Lemma typ_Type : forall e n, typ e (type n) (type (S n)).
red; intros; simpl.
apply (grot_succ_in gr).
Qed.

Lemma cumul_Type : forall e n, sub_typ e (type n) (type (S n)).
red; simpl; intros.
red; intros.
apply ecc_incl; trivial.
Qed.

Lemma cumul_Prop : forall e, sub_typ e prop (type 0).
red; simpl; intros.
red; intros.
apply G_trans with props; trivial.
 apply (grot_succ_typ gr).

 apply (grot_succ_in gr).
Qed.


Lemma typ_prod2 : forall e n T U,
  typ e T (type n) ->
  typ (T :: e) U (type n) ->
  typ e (Prod T U) (type n).
Proof.
unfold typ, Prod; simpl; red; intros e n T U ty_T ty_U i is_val.
apply cc_prod_univ.
 apply ecc_grot.

 red; red; intros.
 rewrite H0; auto with *.

 apply ty_T; trivial.

 intros.
 apply ty_U.
 apply vcons_add_var; auto.
Qed.


(* Ordinals *)

Definition Ord (o:set) := cst o.

Definition typ_ord_kind : forall e o, typ e (Ord o) kind.
red; simpl; intros.
trivial.
Qed.

Definition typ_ord_ord : forall e o o',
  lt o o' -> typ e (Ord o) (Ord o').
red; simpl; intros; trivial.
Qed.


Definition OSucc : term -> term.
intros o; left; exists (fun i => osucc (int o i)).
do 2 red; intros.
rewrite H; reflexivity.
Defined.



(* Nat *)

Definition Nat := cst NAT.

Definition Nati (o:set) := cst (NATi o).

Definition NatI : term -> term.
intros o.
left; exists (fun i => NATi (int o i)).
do 2 red; intros.
apply NATi_morph.
rewrite H; reflexivity.
Defined.

Lemma typ_nati : forall e o,
  isOrd o -> typ e (Nati o) kind.
red; intros; simpl; trivial.
Qed.

Lemma typ_natI : forall e o,
  typ e (NatI o) kind.
red; intros; simpl; trivial.
Qed.

(* Zero *)

Definition Zero := cst ZERO.

Lemma typ_zero : forall e o,
  isOrd o -> typ e Zero (Nati (osucc o)).
red; simpl; intros.
red.
unfold NATi.
apply TI_intro with o; auto.
apply ZERO_typ_gen.
Qed.

(* Successor *)

Definition SuccI : term -> term.
intros o.
apply (Abs (NatI o)).
left; exists (fun i => ZFindtypes.SUCC (i 0)).
do 2 red; intros.
unfold SUCC.
unfold eqX.
apply inr_morph.
apply H.
Defined.


Lemma typ_succi : forall e o i,
  isOrd o ->
  typ e i (Ord o) ->
  typ e (SuccI i) (Prod (NatI i) (NatI (OSucc (lift 1 i)))). 
intros.
unfold SuccI.
apply typ_abs.
 2:discriminate.
red; simpl; intros.
assert (ty_arg :=H1).
red in ty_arg.
specialize ty_arg with (n:=0) (1:=refl_equal _).
simpl in ty_arg.
assert (isOrd (int (lift 1 i) i0)).
 assert (typ (NatI i::e) (lift 1 i) (Ord o)).
  admit. (*weakening*)
 apply H2 in H1.
 simpl in H1.
 apply isOrd_inv with o; auto.
apply TI_intro with (int (lift 1 i) i0); auto.
apply SUCC_typ_gen.
apply eq_elim with (2:=ty_arg).
fold NATi.
apply NATi_morph.
unfold lift; rewrite int_lift_rec_eq; auto with *.
Qed.

(* Case analysis *)

Definition Natcase (fZ fS n : term) : term.
intros fZ fS n.
left; exists (fun i => NATCASE (int fZ i) (fun k => int fS (V.cons k i)) (int n i)).
do 2 red; intros.
red.
apply NATCASE_morph.
 rewrite H; reflexivity.

 red; intros.
 rewrite H; rewrite H0; reflexivity.

 rewrite H; reflexivity.
Defined.

Instance Natcase_morph :
  Proper (eq_term ==> eq_term ==> eq_term ==> eq_term) Natcase.
do 4 red; intros.
simpl; red; intros.
unfold NATCASE.
apply ZFrepl.uchoice_morph_raw.
red; intros.
unfold NATCASE_rel.
apply and_iff_morphism.
 rewrite H; rewrite H1; rewrite H2; rewrite H3; reflexivity.

 apply fa_morph; intros.
 rewrite H0; rewrite H1; rewrite H2; rewrite H3; reflexivity.
Qed.

Lemma Natcase_zero : forall e fZ fS,
  eq_typ e (Natcase fZ fS Zero) fZ.
red; simpl; intros.
red.
apply NATCASE_ZERO.
do 2 red; intros.
rewrite H0; reflexivity.
Qed.

Lemma Natcase_succ : forall o O e n fZ fS,
  isOrd o ->
  typ e O (Ord o) ->
  typ e n (NatI O) ->
  eq_typ e (Natcase fZ fS (App (SuccI O) n)) (subst n fS).
red; intros.
red in H0; specialize H0 with (1:=H2).
red in H1; specialize H1 with (1:=H2).
red; simpl in *.
rewrite <- (fun e1 e2 => NATCASE_morph (int fZ i) (int fZ i) e1
  (fun k => int fS(V.cons k i)) (fun k => int fS(V.cons k i)) e2
  (SUCC (int n i))); auto with *.
 rewrite (NATCASE_SUCC (int O i)); auto.
  rewrite int_subst_eq; reflexivity.

  do 2 red; intros.
  rewrite H3; reflexivity.

  apply isOrd_inv with o; auto.

 red; intros.
 rewrite H3; reflexivity.

 rewrite beta_eq; auto with *.
 red; intros.
 unfold SUCC; apply inr_morph; trivial.
Qed.

Existing Instance NATi_morph.

Lemma typ_natcase : forall o e O P fZ fS n,
  isOrd o ->
  typ e O (Ord o) ->
  typ e fZ (App P Zero) ->
  typ (NatI O :: e) fS (App (lift 1 P) (App (SuccI (lift 1 O)) (Ref 0))) ->
  typ e n (NatI (OSucc O)) ->
  typ e (Natcase fZ fS n) (App P n).
red; intros.
 red in H0; specialize H0 with (1:=H4).
 simpl in H0; red in H0.
 red in H1; specialize H1 with (1:=H4).
 simpl in H1; red in H1.
 red in H3; specialize H3 with (1:=H4).
 simpl in H3; red in H3.
 simpl; red.
 apply NATCASE_typ with (o:=int O i) (P:=fun n => app (int P i) n); trivial.
  do 2 red; intros.
  rewrite H5; reflexivity.

  do 2 red; intros.
  rewrite H5; reflexivity.

  apply isOrd_inv with o; trivial.

  intros.
  assert (val_ok (NatI O :: e) (V.cons n0 i)).
   apply vcons_add_var; trivial.
  apply H2 in H6; clear H1; simpl in H6.
  rewrite beta_eq in H6; trivial.
   rewrite simpl_int_lift1 in H6; trivial.

   red; intros.
   unfold SUCC; apply inr_morph; trivial.
   rewrite simpl_int_lift1; auto.
Qed.

(* Recursor (without case analysis) *)

Require Import ZFfixfuntyp.
Require Import ZFrelations.
Import ModelZF.CCM.

Parameter fextends : set -> set -> set -> Prop.
Parameter fext_morph : Proper (eq_set ==> eq_set ==> eq_set ==> iff) fextends.
Existing Instance fext_morph.

Lemma fext_trans : forall x y f g h,
  fextends x f g -> fextends y g h -> fextends x f h.
Admitted.
(*
Lemma fext_refl : forall f x F,
  f \in prod x F -> fextends x f f.
Admitted.
*)
Parameter extends_app : forall f g x A B,
  f \in prod A B ->
  fextends A f g ->
  x \in A ->
  app f x == app g x.

Parameter fext_intro : forall A B f g,
  f \in prod A B ->
  (forall x, x \in A -> app f x == app g x) ->
  fextends A f g.

Lemma fext_lam : forall A B F f g,
    eq_fun A f f ->
    eq_fun B g g ->
    eq_fun A F F ->
    (forall x, x \in A -> x \in B) ->
    (forall x, x \in A -> f x \in F x) ->
    (forall x, x \in A -> f x == g x) ->
    fextends A (lam A f) (lam B g).
intros.
apply fext_intro with F; intros.
 apply prod_intro; trivial.

 rewrite beta_eq; trivial.
 rewrite beta_eq; auto.
Qed.


Definition fincreasing_bounded o A F :=
  forall x x', lt x' o -> lt x x' -> fextends (A x) (F x) (F x').


Parameter prd_sup : forall o o' A f,
  isOrd o ->
  fincreasing_bounded o A f ->
  lt o' o ->
  fextends (A o') (f o') (sup o f).

Parameter prd_sup_lub : forall o A f g,
  isOrd o ->
  fincreasing_bounded o A f ->
  (forall o', lt o' o -> fextends (A o') (f o') g) ->
  fextends (sup o A) (sup o f) g.

Parameter prd_union : forall o f A B,
  isOrd o ->
  fincreasing_bounded o A f ->
  (forall o', lt o' o -> f o' \in prod (A o') B) ->  
  sup o f \in prod (sup o A) B.

Section NatRec.

  Variable eps : set.
  Hypothesis epsOrd : isOrd eps.
  Variable F : set -> set -> set.
  Hypothesis Fm : morph2 F.
  Variable U : set -> set.
  Hypothesis Um : forall o, lt o eps -> ext_fun (NATi o) U.
  Let Ty o := prod (NATi o) U.
  Hypothesis Ftyp : forall o f, lt o eps ->
    f \in Ty o -> F o f \in Ty (osucc o).

  Definition stab_fix_prop :=
    forall o o' f g,
    lt o' eps -> lt o o' ->
    f \in Ty o -> g \in Ty o' ->
    fextends (NATi o) f g ->
    fextends (NATf (NATi o)) (F o f) (F o' g).

  Hypothesis Fstab : stab_fix_prop.

  Definition NATREC := TIO F.


  Lemma NATREC_inv :
    forall x', lt x' eps ->
    TIO F x' \in Ty x' /\
    forall x, lt x x' ->
    fextends (NATi x) (TIO F x) (TIO F x').
apply isOrd_ind with (3:=epsOrd); intros.
 rewrite <- H1 in H3; auto.

 assert (isOrd x').
  apply isOrd_inv with y; trivial.
 assert (fincreasing_bounded x' (fun z => NATf (NATi z)) (fun z => F z (TIO F z))).
  red; intros.
  destruct H1 with (1:=H2) (2:=H4).
  apply Fstab; auto.
   apply isOrd_trans with x'; trivial.
   apply H0 in H2; trivial.

   destruct H1 with x' x; trivial.
   apply isOrd_trans with x'0; trivial.
 assert (TIO F x' \in Ty x').
  rewrite TIO_eq; auto.
  unfold Ty, NATi.
  setoid_replace (prod (TI NATf x') U) with
    (prod (sup x' (fun o => NATf (TI NATf o))) U).
  2:apply prod_ext; auto with *.
  2:apply TI_eq; auto.
  2:apply Um; apply H0 in H2; auto.
  apply prd_union; trivial.
  intros.
  setoid_replace (prod (NATf (TI NATf o')) U) with
    (prod (TI NATf (osucc o')) U).
  2:symmetry; apply prod_ext; auto with *.
  2:apply TI_mono_succ; auto.
  2:apply isOrd_inv with x'; trivial.
  2:apply Um; apply H0 in H2; auto.
  2:apply le_lt_trans with x'; trivial.
  2:apply lt_osucc_compat; trivial.
  apply Ftyp; auto.
   apply isOrd_trans with x'; trivial.
   apply H0 in H2; trivial.

   destruct H1 with (1:=H2) (2:=H5); trivial.
 split; trivial.
 intros.
 specialize prd_sup with (1:=H3) (2:=H4) (3:=H6); intro.
 rewrite <- TIO_eq in H7; auto.
 apply fext_trans with (2:=H7).
 rewrite TIO_eq; auto.
 2:apply isOrd_inv with x'; auto.
 unfold NATi; rewrite TI_eq; auto.
 2:apply isOrd_inv with x'; auto.
 apply prd_sup_lub.
  apply isOrd_inv with x'; auto.

  red; intros.
  apply H4; trivial.
  apply isOrd_trans with x; trivial.

  intros.
  destruct H1 with (1:=H2) (2:=H6).
  apply Fstab; auto.
   apply H0.
   apply isOrd_trans with x'; trivial.

   destruct H1 with (2:=H8); trivial.
   apply isOrd_trans with x'; trivial.

   rewrite <- TIO_eq; auto.
   apply isOrd_inv with x'; auto.

   rewrite <- TIO_eq; auto.
   apply isOrd_inv with x'; auto.
Qed.

  Lemma NATREC_wt : forall o, lt o eps -> NATREC o \in Ty o.
intros.
destruct NATREC_inv with (1:=H); trivial.
Qed.

  Lemma NATREC_step : forall o,
    lt o eps ->
    fextends (NATi o) (NATREC o) (F o (NATREC o)).
intros.
assert (isOrd o).
 apply isOrd_inv with eps; trivial.
unfold NATREC.
rewrite TIO_eq; auto.
unfold NATi; rewrite TI_eq; auto.
apply prd_sup_lub; intros; auto.
 red; intros.
 apply Fstab; auto.
  apply isOrd_trans with o; trivial.

  apply NATREC_wt.
  apply isOrd_trans with o; trivial.
  apply isOrd_trans with x'; trivial.

  apply NATREC_wt.
  apply isOrd_trans with o; trivial.

  destruct NATREC_inv with x'; auto.
  apply isOrd_trans with o; trivial.

 rewrite <- TIO_eq; auto.
 apply Fstab; auto.
  apply NATREC_wt.
  apply isOrd_trans with o; trivial.

  apply NATREC_wt; trivial.

  destruct NATREC_inv with (1:=H); auto.
Qed.

Section NATREC_Eqn.

  Variable o : set.
  Hypothesis lt_eps : lt o eps.

  Let o_ord : isOrd o.
apply isOrd_inv with eps; trivial.
Qed.

  Lemma NATREC_expand : forall n,
    n \in NATi o -> app (NATREC o) n == app (F o (NATREC o)) n.
intros.
apply extends_app with (NATi o) U; trivial.
 apply NATREC_wt; trivial.

 apply NATREC_step; trivial.
Qed.

  Lemma NATREC_eqn :
    NATREC o == lam (NATi o) (fun x => app (F o (NATREC o)) x).
intros.
rewrite (ZFcoc.cc_eta_eq (NATi o) U (NATREC o)).
2:apply NATREC_wt; auto.
apply lam_ext; auto with *.
red; intros.
rewrite <- H0.
apply NATREC_expand; trivial.
Qed.

End NATREC_Eqn.

End NatRec.

Module Beq.
Definition t := bool.
Definition eq := @eq bool.
Definition eq_equiv : Equivalence eq := eq_equivalence.
Existing Instance eq_equiv.
End Beq.
Module B := VarMap.Make(Beq).

Lemma NATREC_morph :
  forall f1 f2,
  (eq_set ==> eq_set ==> eq_set)%signature f1 f2 ->
  forall o1 o2, isOrd o1 ->
  o1 == o2 ->
  NATREC f1 o1 == NATREC f2 o2.
intros.
unfold NATREC.
unfold TIO.
unfold TR.
apply ZFrepl.uchoice_morph.
 apply TR_rel_choice_pred; auto.
 intros.
 apply sup_morph; auto.
 red; intros.
 assert (morph2 f1).
  do 3 red; intros.
  transitivity (f2 y y0).
   apply H; trivial.

   symmetry; apply H; reflexivity.
 apply H6; trivial.
 apply H3; trivial.

 intros.
 unfold TR_rel.
 split; intros.
  rewrite <- H1.
  apply H2 with (P:=P); intros; auto.
  apply H4 in H6; auto.
  revert H6; apply iff_impl.
  apply H3; auto with *.
  apply sup_morph; auto with *.
  red; intros.
  symmetry; apply H; symmetry; trivial.
  apply H5; trivial.

  rewrite H1.
  apply H2 with (P:=P); intros; auto.
  apply H4 in H6; auto.
  revert H6; apply iff_impl.
  apply H3; auto with *.
  apply sup_morph; auto with *.
  red; intros.
  apply H; trivial.
  apply H5; trivial.
Qed.

(* NatFix o U M is a fixpoint of type Nati o -> U with body M *)
(* How to ensure Fstab ? *)
Definition NatFix (O U M:term) : term.
intros O U M.
left.
exists (fun i =>
  NATREC (fun o' f => int M (V.cons f (V.cons o' i))) (toOrd (int O i))).
red; red; intros.
setoid_replace (toOrd (int O x)) with (toOrd (int O y)).
 apply NATREC_morph.
  do 2 red; intros.
  rewrite H; rewrite H0; rewrite H1; reflexivity.

  apply toOrd_isOrd.

  reflexivity.
  
 rewrite H.
 reflexivity.
Defined.

(* Additional judgments for fix body *)
Definition val_mono (e:env) (doms:env) (ords fixs:B.map) i i' :=
    val_ok e i /\
    val_ok e i' /\
    forall n,
    if ords n then isOrd (i' n) /\ lt (i n) (i' n)
    else if fixs n then
      forall T, nth_error doms n = value T ->
      fextends (int (lift (S n) T) i) (i n) (i' n)
    else i n == i' n.

Definition fx_extends e doms ords fixs dom M :=
    forall i i', val_mono e doms ords fixs i i' ->
    fextends (int dom i) (int M i) (int M i').



Section NatFixRules.

  Variable eps : set.
  Hypothesis eps_ord : isOrd eps.
  Variable e : env.
  Variable O U M : term.
  Hypothesis M_nk : ~ eq_term M kind.
  Hypothesis ty_O : typ e O (Ord eps).
  Hypothesis ty_M : typ (Prod (NatI (Ref 0)) (lift1 1 U)::OSucc O::e)
    M (Prod (NatI (OSucc (Ref 1))) (lift1 2 U)).

  Hypothesis stab : fx_extends (Prod (NatI (Ref 0)) (lift1 1 U) :: OSucc O :: e)
   (NatI (Ref 0) :: nil)
   (B.cons false (B.cons true (B.nil false)))
   (B.cons true (B.cons false (B.nil false)))
   (NatI (OSucc (Ref 1)))
   M.


  Lemma morph_fix_body : forall i M,
    morph2 (fun o' f => int M (V.cons f (V.cons o' i))).
do 3 red; intros.
rewrite H; rewrite H0; reflexivity.
Qed.

  Lemma ext_fun_ty : forall o i,
    ext_fun (NATi o) (fun x => int U (V.cons x i)).
do 2 red; intros.
rewrite H0;reflexivity.
Qed.

  Lemma val_M_ok : forall i o f,
    val_ok e i ->
    lt o (osucc (int O i)) ->
    f \in prod (NATi o) (fun x => int U (V.cons x i)) ->
    val_ok (Prod (NatI (Ref 0)) (lift1 1 U) :: OSucc O :: e)
      (V.cons f (V.cons o i)).
intros.
repeat apply vcons_add_var; trivial.
simpl.
apply eq_elim with (2:=H1).
apply prod_ext.
 reflexivity.

 red; intros.
 rewrite simpl_lift1.
 rewrite lift10.
 rewrite H3; reflexivity.
Qed.

  Lemma ty_fix_body : forall i o f,
   val_ok e i ->
   lt o (osucc (int O i)) ->
   f \in prod (NATi o) (fun x => int U (V.cons x i)) ->
   int M (V.cons f (V.cons o i)) \in
   prod (NATi (osucc o)) (fun x => int U (V.cons x i)).
intros.
red in ty_M.
specialize ty_M with (i:=V.cons f (V.cons o i)).
simpl in ty_M.
setoid_replace (prod (NATi (osucc o)) (fun x => int U (V.cons x i)))
   with (prod (NATi (osucc o))
          (fun x => int (lift1 2 U) (V.cons x (V.cons f (V.cons o i))))).
 apply ty_M.
 apply val_M_ok; trivial.

 apply prod_ext; auto with *; intros.
 red; intros.
 do 2 rewrite simpl_lift1; rewrite lift10.
 rewrite H3; reflexivity.
Qed.


  Lemma stab_fix_body : forall i,
    val_ok e i ->
    stab_fix_prop (osucc (int O i))
      (fun o' f => int M (V.cons f (V.cons o' i)))
      (fun x => int U (V.cons x i)).
red; intros.
assert (isOrd (int O i)).
 apply ty_O in H.
 apply isOrd_inv with eps; auto.
assert (isOrd o').
 apply isOrd_inv with (2:=H0); auto.
red in stab.
unfold NATi; rewrite <- TI_mono_succ; auto. 
2:apply isOrd_inv with o'; trivial.
assert (Hstab := stab (V.cons f (V.cons o i)) (V.cons g (V.cons o' i))).
simpl in Hstab.
apply Hstab.
split.
 apply val_M_ok; auto.
 apply isOrd_trans with o'; auto.
split.
 apply val_M_ok; auto.
destruct n as [|[|n]]; simpl.
 intros.

 injection H7; clear H7; intros; subst T.
 simpl.
 unfold V.lams, V.shift; simpl; trivial.

 split; trivial.

 reflexivity.
Qed.


Lemma typ_nat_fix :
  typ e (NatFix O U M) (Prod (NatI O) U).
red; intros.
simpl.
assert (isOrd (int O i)).
 apply ty_O in H.
 simpl in H.
 apply isOrd_inv with eps; trivial.
unfold NATi; rewrite toOrd_ord; trivial.
apply NATREC_wt with (eps :=osucc(int O i)); intros; auto.
 apply morph_fix_body.

 do 2 red; intros.
 rewrite H3; reflexivity.

 apply ty_fix_body; trivial.

 apply stab_fix_body; trivial.
Qed.

Lemma nat_fix_eq : forall N,
  typ e N (NatI O) ->
  eq_typ e (App (NatFix O U M) N)
           (App (subst O (subst (lift 1 (NatFix O U M)) M)) N).
red; intros.
generalize ty_fix_body; intro tfb.
generalize stab_fix_body; intro sfb.
change
 (app (NATREC (fun o f => int M (V.cons f (V.cons o i))) (toOrd (int O i))) (int N i) ==
  app (int (subst O (subst (lift 1 (NatFix O U M)) M)) i) (int N i)).
do 2 rewrite int_subst_eq.
rewrite simpl_int_lift.
rewrite lift0_term.
simpl.
assert (O_lt := @ty_O _ H0).
simpl in O_lt.
specialize NATREC_expand
 with (1:=isOrd_succ _ (isOrd_inv _ _ eps_ord O_lt)) (2:=@ty_O _ H0); intro.
rewrite H1. (* with (U:=fun x => int U (V.cons x i)) *)
 apply app_ext; try reflexivity.
 apply int_morph; auto with *.
 apply V.cons_morph.
  apply NATREC_morph; auto with *.
   do 2 red; intros.
   apply int_morph; auto with *.
   apply V.cons_morph; trivial.
   rewrite H2.
   reflexivity.

   apply toOrd_isOrd.

   rewrite toOrd_ord; auto with *.
   apply ty_O in H0.
   apply isOrd_inv with eps; trivial.

 do 3 red; intros.
 rewrite H2; rewrite H3; reflexivity.

 2:intros.
 2:apply tfb; trivial.

 do 2 red; intros.
 rewrite H4; reflexivity.

 intros.
 apply sfb; trivial.

 rewrite toOrd_ord.
  apply lt_osucc.
  apply isOrd_inv with eps; trivial.

  apply isOrd_inv with eps; trivial.

 rewrite toOrd_ord.
  apply H in H0; trivial.

  apply isOrd_inv with eps; trivial.
Qed.
 
End NatFixRules.

(* More judgments *)

  Definition noccur (f:nat->bool) (T:term) : Prop :=
    forall i i',
    (forall n, if f n then True else i n == i' n) ->
    int T i == int T i'.

  Lemma noc_var : forall f n, f n = false -> noccur f (Ref n).
red; simpl; intros.
specialize (H0 n).
rewrite H in H0; trivial.
Qed.

  Lemma noc_kind : forall f, noccur f kind.
red; simpl; reflexivity.
Qed.

  Lemma noc_prop : forall f, noccur f prop.
red; simpl; reflexivity.
Qed.

  Lemma noc_app : forall f a b,
    noccur f a -> noccur f b -> noccur f (App a b).
unfold noccur; simpl; intros.
rewrite (H _ _ H1).
rewrite (H0 _ _ H1).
reflexivity.
Qed.



  Definition fx_equals e doms ords fixs M :=
    forall i i', val_mono e doms ords fixs i i' ->
    int M i == int M i'.

  Definition fx_sub e doms ords fixs M :=
    forall i i', val_mono e doms ords fixs i i' ->
    int M i \incl int M i'.

  Lemma fx_abs : forall e doms ords fixs U T T' M,
    fx_sub e doms ords fixs T ->
    fx_equals (T::e) (T'::doms) (B.cons false ords) (B.cons false fixs) M ->
    typ (T::e) M U ->
    U <> kind ->
    fx_extends e doms ords fixs T (Abs T M).
unfold fx_sub, fx_equals, fx_extends; intros.
specialize H with (1:=H3).
simpl.
apply fext_lam with (F:= fun x => int U (V.cons x i)).
 red; intros.
 rewrite H5; reflexivity.

 red; intros.
 rewrite H5; reflexivity.

 red; intros.
 rewrite H5; reflexivity.

 intros.
 apply H; trivial.

 intros.
 red in H1.
 destruct U as [(U0,Um)|]; simpl in *; try (elim H2; reflexivity).
 apply H1.
 apply vcons_add_var; trivial.
 destruct H3; trivial.

 intros.
 apply H0.
 destruct H3.
 destruct H5.
 split;[|split].
  apply vcons_add_var; auto.

  apply vcons_add_var; auto.
  apply H; trivial.

  destruct n as [|n]; simpl; auto with *.

  generalize (H6 n).
  destruct (ords n); auto.
  destruct (fixs n); intros; auto.
  rewrite simpl_int_lift; auto.
Qed.

  Lemma NATi_sub : forall e doms ords fixs o O,
    isOrd o ->
    typ e O (Ord o) ->
    fx_sub e doms ords fixs O ->
    fx_sub e doms ords fixs (NatI O).
unfold fx_sub; intros.
simpl.
apply TI_mono; auto.
 destruct H2 as (_,(H2,_)).
 apply H0 in H2.
 apply isOrd_inv with o; auto.

 destruct H2 as (H2,_).
 apply H0 in H2.
 apply isOrd_inv with o; auto.
Qed.

  Lemma OSucc_sub : forall e doms ords fixs O,
    fx_sub e doms ords fixs O ->
    fx_sub e doms ords fixs (OSucc O).
unfold fx_sub; simpl; intros.
red; intros.
unfold osucc in *.
apply subset_intro.
 apply power_intro; intros.
 apply H with (1:=H0); trivial.
 apply power_elim with z; trivial.
 apply subset_elim1 in H1; trivial.

 apply subset_elim2 in H1.
 destruct H1.
 rewrite H1; trivial.
Qed.

  Lemma var_sub : forall e doms ords fixs n,
    ords n = true ->
    fx_sub e doms ords fixs (Ref n).
unfold fx_sub; intros.
destruct H0 as (_,(_,H0)).
generalize (H0 n); rewrite H.
destruct 1; simpl.
red; intros.
apply isOrd_trans with (i n); auto.
Qed.

  Lemma fx_eq_app : forall e doms ords fixs u v,
    fx_equals e doms ords fixs u ->
    fx_equals e doms ords fixs v ->
    fx_equals e doms ords fixs (App u v).
unfold fx_equals; intros.
specialize H with (1:=H1).
specialize H0 with (1:=H1).
simpl.
rewrite H; rewrite H0; reflexivity.
Qed.

Import Bool.
  Lemma fx_eq_noc : forall e doms ords fixs t,
    noccur (fun n => ords n || fixs n) t ->
    fx_equals e doms ords fixs t.
unfold noccur, fx_equals; intros.
apply H; intros.
destruct H0 as (_,(_,H0)).
generalize (H0 n).
destruct (ords n); simpl; trivial.
destruct (fixs n); simpl; trivial.
Qed.

  Lemma fx_eq_rec_call : forall e doms ords fixs n x T U,
    ords n = false ->
    fixs n = true ->
    nth_error doms n = value T ->
    T <> kind ->
    typ e (Ref n) (Prod (lift (S n) T) U) ->
    fx_equals e doms ords fixs x ->
    typ e x (lift (S n) T) ->
    fx_equals e doms ords fixs (App (Ref n) x).
unfold fx_equals; intros.
simpl.
specialize H4 with (1:=H6).
rewrite <- H4.
destruct H6 as (Hty,(_,Hrec)).
specialize Hrec with n.
rewrite H in Hrec; rewrite H0 in Hrec.
apply Hrec in H1.
apply extends_app with (int (lift (S n) T) i) (fun x => int U (V.cons x i)); trivial.
 red in H3; specialize H3 with (1:=Hty); trivial.

 red in H5; specialize H5 with (1:=Hty); trivial.
 destruct T as [(T,Tm)|]; simpl in *; auto.
 elim H2; trivial.
Qed.


Section Example.

Definition nat_ind_typ :=
   Prod (Prod (NatI (Ord omega)) prop) (* P : nat -> Prop *)
  (Prod (App (Ref 0) Zero)
  (Prod (Prod (NatI (Ord omega)) (Prod (App (Ref 2) (Ref 0))
                        (App (Ref 3) (App (SuccI (Ord omega)) (Ref 1)))))
  (Prod (NatI (Ord omega)) (App (Ref 3) (Ref 0))))).

Definition nat_ind :=
   Abs (*P*)(Prod (NatI (Ord omega)) prop) (* P : nat -> Prop *)
  (Abs (*fZ*) (App (Ref 0) Zero)
  (Abs (*fS*) (Prod (*n*)(NatI (Ord omega)) (Prod (App (Ref 2) (Ref 0))
                                   (App (Ref 3) (App (SuccI (Ord omega)) (Ref 1)))))
  (NatFix (Ord omega) (*n*)(App (Ref 3) (Ref 0))
    (*o,Hrec*)
    (Abs (*n*)(NatI (OSucc (Ref 1)))
      (Natcase
        (Ref 4)
        (*k*)(App (App (Ref 4) (Ref 0))
                  (App (Ref 2) (Ref 0)))
        (Ref 0)))))).

Lemma typ_var' : forall e n T,
  match nth_error e n with value T' => eq_term (lift (S n) T') T | _ => False end ->
  typ e (Ref n) T.
intros.
case_eq (nth_error e n); intros.
 rewrite H0 in H.
 apply typ_var in H0.
 revert H0; apply iff_impl.
 apply typ_morph; auto with *.

 rewrite H0 in H; contradiction.
Qed.

Lemma typ_app0 : forall e u v U V Ur,
  typ e v V ->
  typ e u (Prod V Ur) ->
  V <> kind ->
  Ur <> kind ->
  sub_typ e (subst v Ur) U ->
  typ e (App u v) U.
intros.
apply typ_subsumption with (subst v Ur); trivial.
 apply typ_app with V; trivial.

 destruct Ur as [(Ur,Urm)|]; simpl; trivial; discriminate.
Qed.

Lemma typ_app' : forall e u v U V Ur,
  typ e v V ->
  typ e u (Prod V Ur) ->
  V <> kind ->
  eq_term U (subst v Ur) ->
  typ e (App u v) U.
intros.
assert (typ e (App u v) (subst v Ur)).
 apply typ_app with V; trivial.
revert H3; apply iff_impl.
apply typ_morph; auto with *.
Qed.

Lemma eqterm_subst_App : forall N u v,
  eq_term (subst N (App u v)) (App (subst N u) (subst N v)).
red; simpl; intros.
red; intros.
unfold subst.
do 2 rewrite int_subst_rec_eq.
rewrite H.
reflexivity.
Qed.

Lemma nat_ind_def :
  forall e, typ e nat_ind nat_ind_typ.
assert (eq_term Nat (NatI (Ord omega))).
 red; simpl.
 red; unfold NAT; reflexivity.
unfold nat_ind, nat_ind_typ; intros.
apply typ_abs; try discriminate.
apply typ_abs; try discriminate.
apply typ_abs; try discriminate.
set (E :=NatI (OSucc (Ref 1))
      :: Prod (NatI (Ref 0)) (lift1 1 (App (Ref 3) (Ref 0)))
         :: OSucc (Ord omega)
            :: Prod (NatI (Ord omega))
                 (Prod (App (Ref 2) (Ref 0))
                    (App (Ref 3) (App (SuccI (Ord omega)) (Ref 1))))
               :: App (Ref 0) Zero :: Prod (NatI (Ord omega)) prop :: e).
assert (ty_case :
   typ E
     (Natcase (Ref 4) (App (App (Ref 4) (Ref 0)) (App (Ref 2) (Ref 0)))
        (Ref 0)) (App (Ref 5) (Ref 0))).
  apply typ_natcase with (o:=osucc omega) (O0:=Ref 2); auto.
   apply typ_var'; simpl.
   repeat red; simpl; intros.
   reflexivity.

   apply typ_var'; simpl.
   repeat red; simpl; intros.
   unfold V.lams, V.shift; simpl.
   apply app_ext; auto with *.

   apply typ_app0 with (App (Ref 6) (Ref 0))
     (App (Ref 7) (App (SuccI (Ord omega)) (Ref 1))); try discriminate.
    apply typ_app' with (NatI (Ref 3)) (App (Ref 7) (Ref 0)); try discriminate.
     apply typ_var'; simpl; auto with *.
     repeat red; intros.
     unfold V.lams, V.shift; simpl.
     apply NATi_morph; apply H0.

     apply typ_var'; simpl.
     repeat red; intros.
     apply prod_ext.
      unfold V.lams, V.shift; simpl.
      apply NATi_morph; apply H0.

      red; intros.
      unfold V.lams, V.shift; simpl.
      apply app_ext; auto with *.

      simpl.
      repeat red; intros.
      unfold V.lams, V.shift; simpl.
      apply app_ext; auto with *.

    apply typ_app' with Nat (Prod (App (Ref 7) (Ref 0))
        (App (Ref 8) (App (SuccI (Ord omega)) (Ref 1)))); try discriminate.
     apply typ_subsumption with (NatI (Ref 3)); try discriminate.
      apply typ_var'; simpl; auto.
      repeat red; intros.
      unfold V.lams, V.shift; simpl.
      apply NATi_morph; apply H0.

      red; simpl; intros.
      apply NATi_NAT with (o:=i 3); trivial.
      red in H0.
      specialize (H0 3 _ (eq_refl _)).
      simpl in H0.
      apply isOrd_inv with (osucc omega); auto.

      apply typ_var'; simpl; auto.
      repeat red; intros.
      apply prod_ext.
       reflexivity.
      red; intros.
      apply prod_ext; auto with *.
       unfold V.lams, V.shift; simpl.
       apply app_ext; auto with *.

       red; intros.
       apply app_ext; auto with *.
        unfold V.lams, V.shift; simpl; auto.

        apply app_ext; auto with *.

      simpl; repeat red; intros.
      apply prod_ext; auto with *.
       unfold V.lams, V.shift; simpl.
       apply app_ext; auto.

       red; intros.
       apply app_ext; auto.
        unfold V.lams, V.shift; simpl; auto.

        apply app_ext; auto with *.
        unfold V.lams, V.shift; simpl; auto.

    repeat rewrite eqterm_subst_App.
    apply sub_refl.
    apply eq_typ_app.
     red; simpl.
     unfold V.lams, V.shift; simpl; reflexivity.
    red; simpl; intros.
    rewrite beta_eq.
     rewrite beta_eq.
      unfold V.lams, V.shift; simpl.
      reflexivity.

      red; intros.
      rewrite H2; reflexivity.

      unfold V.lams, V.shift; simpl.
      specialize (H0 0 _ (eq_refl _)); simpl in H0.
      exact H0.

     red; intros.
     rewrite H2; reflexivity.

     unfold V.lams, V.shift; simpl.
     generalize (H0 0 _ (eq_refl _)); intro H1; simpl in H1.
     apply NATi_NAT with (o:=i 3); try assumption.
     specialize (H0 3 _ (eq_refl _)); simpl in H0.
     apply isOrd_inv with (osucc omega); auto.

   apply typ_var'; simpl.
   red; intros.
   unfold V.lams, V.shift; simpl.
   rewrite (H0 2); reflexivity.

apply typ_nat_fix with (eps:=osucc omega); auto.
 apply typ_ord_ord; auto.

 apply typ_abs; try discriminate.
 revert ty_case; apply iff_impl.
 apply typ_morph; auto with *.
 repeat red; simpl; intros.
 unfold V.lams, V.shift; simpl.
 apply app_ext; auto with *.

 (* *)
(* apply fx_abs with (U:=App (Ref 5) (Ref 0)) (T':=prop); trivial; try discriminate.
  apply NATi_sub with (o:=osucc (osucc omega)); auto.
   red; intros; simpl.
   generalize (H0 1 _ (eq_refl _)); simpl; intros; clear H0.
   apply lt_osucc_compat; auto.

   apply OSucc_sub.
   apply var_sub; reflexivity.

  fold E.
  set (fxeq := fx_equals E (prop :: NatI (Ref 0) :: nil)
     (B.cons false (B.cons false (B.cons true (B.nil false))))
     (B.cons false (B.cons true (B.cons false (B.nil false))))) in |-*.
  assert (fxeq (Ref 4)).
   apply fx_eq_noc.
   apply noc_var; simpl; reflexivity.
  assert (fxeq (Ref 0)).
   apply fx_eq_noc.
   apply noc_var; simpl; reflexivity.
  assert (fx_equals (NatI (Ref 2) :: E) (prop :: prop :: NatI (Ref 0) :: nil)
            (B.cons false (B.cons false (B.cons false (B.cons true (B.nil false)))))
            (B.cons false (B.cons false (B.cons true (B.cons false (B.nil false)))))
            (App (App (Ref 4) (Ref 0)) (App (Ref 2) (Ref 0)))).
   apply fx_eq_app.
    apply fx_eq_noc; apply noc_app; apply noc_var; simpl; reflexivity.

    apply fx_eq_rec_call with (NatI (Ref 0)) (App (Ref 7) (Ref 0)); simpl; trivial.
     discriminate.

     apply typ_var'; simpl.
     red; intros.
     apply prod_ext.
      unfold V.lams, V.shift; simpl.
      rewrite (H2 3); reflexivity.

      red; intros.
      unfold V.lams, V.shift; simpl.
      rewrite (H2 6); rewrite H4; reflexivity.

     apply fx_eq_noc; apply noc_var; simpl; trivial.

     apply typ_var'; simpl.
     red; intros.
     unfold V.lams, V.shift; simpl.
     rewrite (H2 3); reflexivity.
*)
 red; intros.
 simpl.
 destruct H0.
 destruct H1.
 generalize (H2 1); simpl; destruct 1.
 apply fext_lam with (F:=fun x => app (i 4) x).
  red; intros.
  apply NATCASE_morph; auto with *.
  red; intros.
  rewrite H7; reflexivity.

  red; intros.
  apply NATCASE_morph; auto with *.
  red; intros.
  rewrite H7; reflexivity.

  red; intros.
  rewrite H6; reflexivity.

  unfold NATi; apply TI_incl; auto.
  apply lt_osucc_compat; auto.

  (* typage (again!) *)
  intros.
  generalize (ty_case (V.cons x i)); simpl; intros.
  apply H6.
  apply vcons_add_var; trivial.

  (* invariance: *)
  intros.
  unfold NATCASE.
  apply ZFrepl.uchoice_morph_raw.
red; intros.
unfold NATCASE_rel.
apply and_iff_morphism.
 generalize (H2 3); simpl; intro.
 rewrite H6; rewrite H7; reflexivity.

 apply fa_morph; intro.
 apply fa_morph; intro.
 generalize (H2 2); simpl; intro.
 rewrite H6; rewrite H7.
 setoid_replace (app (i 0) x1) with (app (i' 0) x1).
 reflexivity.

  apply extends_app with (NATi (i 1)) (fun x => app (i 4) x).
   generalize (H0 0 _ (eq_refl _)); simpl; intro.
   apply eq_elim with (2:=H8).
   apply prod_ext; simpl.
    reflexivity.

    red; intros.
    unfold V.lams, V.shift; simpl.
    rewrite H10; reflexivity.

   generalize (H2 0 _ (eq_refl _)); simpl; intro.
   exact H8.

  unfold NATi in H5; rewrite TI_mono_succ in H5; auto.
  2:apply isOrd_inv with (i' 1); auto.
  apply NATf_case with (3:=H5); intros.
   rewrite H8 in x2.
   apply NATf_discr in x2; contradiction.

   rewrite H9 in x2; apply SUCC_inj in x2.
   rewrite <- x2; trivial.
Qed.

End Example.