Require Import ZF.
Require Import ZFrelations.
Import IZF.

Require Import ZFord.
Require Import ZFnats.

Section IterMonotone.

  Variable F : set -> set.
  Variable Fmono : Proper (incl_set ==> incl_set) F.

  Instance Fmono_morph: Proper (eq_set ==> eq_set) F.
do 2 red; intros.
apply eq_intro; intros.
 apply Fmono with x; trivial.
 red; intros.
 rewrite <- H; trivial.

 apply Fmono with y; trivial.
 red; intros.
 rewrite H; trivial.
Qed.
Hint Resolve Fmono_morph.

  Lemma TI_mono : increasing (TI F).
do 2 red; intros.
apply TI_elim in H2; intros; auto with *.
destruct H2.
apply TI_intro with x0; auto with *.
apply H1 in H2; trivial.
Qed.

  Lemma TI_incl : forall o, isOrd o ->
    forall o', lt o' o ->
    TI F o' \incl TI F o.
intros.
apply TI_mono; trivial; auto.
 apply isOrd_inv with o; trivial.
 red; intros; apply isOrd_trans with o'; trivial.
Qed.

  Lemma TI_mono_succ : forall o,
    isOrd o ->
    TI F (osucc o) == F (TI F o).
intros.
assert (Fext : ext_fun (osucc o) (fun o' => F (TI F o'))).
 generalize (isOrd_succ _ H); auto.
rewrite TI_eq; auto.
 apply eq_intro; intros.
  rewrite sup_ax in H0; trivial.
  destruct H0.
  apply Fmono with (TI F x); trivial.
  apply TI_mono; trivial.
   apply isOrd_inv with (osucc o); auto.
   apply olts_le; trivial.

 rewrite sup_ax; trivial.
 exists o; trivial.
 apply lt_osucc; trivial.
Qed.

  Lemma TI_pre_fix : forall fx o,
     isOrd o ->
     F fx \incl fx ->
     TI F o \incl fx.
intros.
induction H using isOrd_ind; intros.
 specialize (TI_morph F _ _ H2); intro.
 rewrite <- H3; trivial.

 red; intros.
 apply H0.
 elim TI_elim with (3:=H3); intros; auto with *.
 apply Fmono with (TI F x); auto.
Qed.

 (* m.q on atteint un pt fix pour un ord assez grand... *)

End IterMonotone.

  (* Von Neumann universes *)
  Definition VN := TI power.

  Lemma VN_trans : forall o x y,
    isOrd o ->
    x \in VN o ->
    y \in x ->
    y \in VN o.
unfold VN; intros.
apply TI_elim in H0; auto with *.
destruct H0.
apply power_elim with (z:=y) in H2; trivial.
apply TI_incl with x0; auto with *.
exact power_mono.
Qed.

  Lemma VN_incl : forall o x y,
    isOrd o ->
    y \incl x ->
    x \in VN o ->
    y \in VN o.
intros.
apply TI_elim in H1; auto with *.
destruct H1.
apply TI_intro with x0; auto with *.
apply power_intro; intros.
apply H0 in H3.
eapply power_elim; eauto.
Qed.

  Lemma VNsucc_power : forall o x,
    isOrd o ->
    x \in VN o ->
    power x \in VN (osucc o).
induction 1 using isOrd_ind; intros.
 setoid_replace (VN (osucc y')) with (VN (osucc y)); auto.
  apply IHisOrd.
  setoid_replace (VN y) with (VN y'); auto.
  unfold VN.
  apply TI_morph; trivial.
 unfold VN.
 apply TI_morph; rewrite <- H1; reflexivity.

 unfold VN in H2 |- *.
 apply TI_elim in H2; auto with *; destruct H2.
 rewrite power_ax in H3.
 apply TI_intro with y; auto with *.
 apply power_intro; intros.
 rewrite power_ax in H4.
 apply TI_intro with x0; auto with *.
 apply power_intro; intros; auto.
Qed.


  Lemma VNlim_clos_power : forall o x, limitOrd o -> x \in VN o -> power x \in VN o.
unfold VN; intros.
apply TI_elim in H0; auto with *.
destruct H0.
apply TI_incl with (osucc (osucc x0)); auto with *.
 exact power_mono.
 destruct H; auto.

apply VNsucc_power.
 apply isOrd_succ; apply isOrd_inv with o; trivial.
 destruct H; trivial.

 unfold VN.
 apply TI_intro with x0; auto with *.
  apply isOrd_succ; apply isOrd_inv with o; trivial.
  destruct H; trivial.

  apply lt_osucc.
  apply isOrd_inv with o; trivial.
  destruct H; trivial.
Qed.

  Lemma VN_clos_union : forall o x, isOrd o -> x \in VN o -> union x \in VN o.
unfold VN; intros.
apply TI_elim in H0; auto with *.
destruct H0.
apply TI_intro with x0; intros; auto with *.
apply power_intro; intros.
apply union_elim in H2; destruct H2.
rewrite power_ax in H1.
apply H1 in H3.
apply TI_elim in H3; auto with *.
2:apply isOrd_inv with o; auto.
destruct H3.
rewrite power_ax in H4.
apply H4 in H2.
apply TI_incl with x2; auto with *.
 exact power_mono.
 apply isOrd_inv with o; auto.
Qed.

Require Import ZFcard.
Require Import ZFrepl.
Require Import ZFordcl ZFord_equiv.
Import ZFord.

  Lemma repl_rel_incl : forall x y R,
    x \incl y ->
    repl_rel y R ->
    repl_rel x R.
destruct 2.
split; intros; eauto.
Qed.

Section Inaccessible.

  Variable mu : set.
  Variable mu_ord : isOrd mu.
  Variable mu_lim : forall x, lt x mu -> lt (osucc x) mu.
  Variable mu_regular : regular mu.

  Lemma VNcard : forall x,
    x \in VN mu -> lt_card x mu.
Admitted.

  Lemma VNlim_clos_union_repl : forall I R,
    repl_rel I R ->
    I \in VN mu ->
    (forall x y, x \in I -> R x y -> y \in VN mu) ->
    union (repl I R) \in VN mu.
unfold VN; intros.
pose (I' := subset I (fun x => exists y, R x y)).
assert (rrI' : repl_rel I' R).
 apply repl_rel_incl with I; trivial.
 red; intros.
 apply subset_elim1 in H2; trivial.
assert (repl I R == repl I' R).
 apply eq_intro; intros.
  apply repl_elim in H2; auto.
  destruct H2.
  apply repl_intro with x; auto.
  apply subset_intro; auto.
  exists z; trivial.

  apply repl_mono2 with I'; auto.
  red; intros.
  apply subset_elim1 in H3; trivial.
rewrite H2; clear H2.
assert (stgI' : I' \in VN mu).
 apply VN_incl with I; trivial; red; intros.
 apply subset_elim1 in H2; trivial.
clear H0.
pose (G := fun x => least_ord mu
   (fun n => forall y, R x y -> y \in VN n)).
assert (Gm : ext_fun I' G).
 unfold G, VN; do 2 red; intros.
 apply least_ord_morph; intros; auto; try reflexivity.
 assert (TI power x0 == TI power x'0).
  apply TI_morph; trivial.
 split; intros.
  rewrite <- H5.
  apply H6.
  destruct H.
  apply H with x' y; trivial.
   apply subset_elim1 in H0.
   rewrite <- H2; trivial.

   symmetry; trivial.
   reflexivity.
   
  rewrite H5.
  apply H6.
  destruct H.
  apply H with x y; trivial.
   apply subset_elim1 in H0; trivial.
   reflexivity.
apply TI_elim in stgI'; auto with *.
destruct stgI'.
assert (Fprop : forall x, x \in I' ->
  (forall y, R x y -> y \in VN (G x)) /\ lt (G x) mu).
 intros.
 assert (exists y, R x0 y).
  specialize subset_elim2 with (1:=H3); destruct 1.
  destruct H5.
  exists x2.
  apply H with (4:=H5); auto.
   apply subset_elim1 in H3; trivial.
   rewrite <- H4; trivial.
   symmetry; trivial.
   reflexivity.
 apply subset_elim1 in H3.
 destruct H4.
 specialize H1 with (1:=H3) (2:=H4).
 apply TI_elim in H1; auto with *.
 destruct H1.
 rewrite <- TI_mono_succ in H5; auto with *.
 2:exact power_mono.
 2:apply isOrd_inv with mu; trivial.
 destruct least_ord1 with
    (o:=mu) (P:=fun n => forall y, R x0 y -> y \in VN n)
    (x:=osucc x2); intros; auto.
  setoid_replace (VN x') with (VN x3); auto.
  symmetry; unfold VN; apply TI_morph; trivial.

  setoid_replace y with x1; auto.
  destruct H; eauto.

  destruct H7.
  destruct H8.
  split; eauto.
  apply le_lt_trans with (osucc x2); auto.
  red in H8; rewrite succ_equiv in H8; auto.
  apply ZFord.isOrd_succ.
  apply ZFord.isOrd_inv with mu; trivial.
assert (Fmu := fun x h => proj2 (Fprop x h)).
assert (Fspec := fun x h => proj1 (Fprop x h)).
clear Fprop.
assert (Ford : forall x, x \in I' -> isOrd (G x)).
 intros.
 apply isOrd_inv with mu; auto.
assert (lt (sup I' G) mu).
 apply mu_regular; intros; auto.
 apply VNcard.
 apply TI_intro with x; auto.
 exact power_morph.
apply TI_intro with (sup I' G); auto with *.
apply power_intro; intros.
apply union_elim in H4; destruct H4.
apply repl_elim in H5; trivial.
destruct H5.
apply VN_trans with x0; auto.
 apply isOrd_inv with mu; auto.

 apply TI_mono with (G x1); eauto.
  exact power_mono.
  apply isOrd_inv with mu; auto.
Qed.


End Inaccessible.

Section InaccessibleCard.

  Variable mu : set.
  Variable mu_card : isCard mu.
  Variable mu_lim : forall x, lt x mu -> lt (osucc x) mu.
  Variable mu_regular : regular_card mu.

  Let mu_ord : isOrd mu.
destruct mu_card; trivial.
Qed.
  Let mu_lt : forall x, lt x mu -> lt_card x mu.
destruct mu_card; trivial.
Qed.

  Lemma lt_card_lt : forall x, isOrd x -> lt_card x mu -> lt x mu.
unfold lt_card; intros.
destruct (ZFordcl.ClassicOrdinal.ord_total mu x); trivial.
 apply isOrd_eqv1; auto.
 apply isOrd_eqv1; auto.

 elim H0.
 apply ZFordcl.ord_le_incl in H1; try (apply isOrd_eqv1; auto; fail).
 apply incl_le_card in H1; trivial.
Qed.


  Lemma VNcard' : forall x,
    x \in VN mu -> lt_card x mu.
Admitted.

  Lemma VNreg_clos_sup : forall I F,
    ext_fun I F ->
    I \in VN mu ->
    (forall x, x \in I -> F x \in VN mu) ->
    sup I F \in VN mu.
unfold VN; intros.
pose (G := fun x => least_ord mu (fun n => F x \in VN n)).
assert (Gm : ext_fun I G).
 unfold G, VN; do 2 red; intros.
 apply least_ord_morph; intros; trivial; try reflexivity.
  apply isOrd_eqv1; trivial.
 assert (TI power x0 == TI power x'0).
  apply TI_morph; trivial.
 rewrite <- H6.
 rewrite (H x x'); trivial.
 reflexivity.
apply TI_elim in H0; auto with *.
destruct H0.
assert (Fprop : forall x, x \in I -> F x \in VN (G x) /\ lt (G x) mu).
 intros.
 specialize H1 with (1:=H3).
 apply TI_elim in H1; auto with *.
 destruct H1.
 rewrite <- TI_mono_succ in H4; auto with *.
 2:exact power_mono.
 2:apply isOrd_inv with mu; trivial.
 destruct least_ord1 with
    (o:=mu) (P:=fun n => F x0 \in VN n)
    (x:=osucc x1); intros; auto.
  setoid_replace (VN x') with (VN x2); auto.
  symmetry; unfold VN; apply TI_morph; trivial.

  destruct H6.
  destruct H7.
  split; eauto.
  apply le_lt_trans with (osucc x1); auto.
  red in H7; rewrite succ_equiv in H7; auto.
  apply ZFord.isOrd_succ.
  apply ZFord.isOrd_inv with mu; trivial.
assert (Fmu := fun x h => proj2 (Fprop x h)).
assert (Fspec := fun x h => proj1 (Fprop x h)).
clear Fprop.
assert (Ford : forall x, x \in I -> isOrd (G x)).
 intros.
 apply isOrd_inv with mu; auto.
assert (sup_ord : isOrd (sup I G)).
 apply isOrd_supf; trivial.
assert (lt (sup I G) mu).
 apply lt_card_lt; trivial.
 apply mu_regular; intros; auto.
 apply VNcard'.
 apply TI_intro with x; auto with *.
apply TI_intro with (sup I G); auto with *.
apply power_intro; intros.
rewrite sup_ax in H4; auto.
destruct H4.
apply VN_trans with (F x0); auto.
apply TI_mono with (G x0); eauto.
exact power_mono.
Qed.


End InaccessibleCard.
